/***************************************************************
L'Automate gère les accelerations, les ralentissements et les arrêts en gare
selon des intervalles de temps d'une seconde et les états de circulation ci-dessous.
Automate::loop est invoqué dans la loop principale pour faire ce qu'il doit faire
Les détecteurs des zones ci-dessous changent l'état de l'automate lors des entrées dans les zones
Il ne peut y avoir qu'une seule zone détectée à la fois donnée par ZoneMonitor
Chaque détection de zone (selon ZoneMonitor::checkZone()) entraine un traitement spécifique par zone
dans la loop principale:
- En gare1: décompte de l'arrêt en gare et changement de sens
- En zone A, dans le sens A->B: accélération 
  et dans le sens B->A: mesure de la vitesse et ralentissement 
- En zone B: initialisation du compteur de vitesse
- En zone C, dans le sens C->B: accélération 
  et dans le sens B->C: mesure de la vitesse et ralentissement 
- En gare2: décompte de l'arrêt en gare et changement de sens
A chaque mesure, la vitesse commandée est modifiée si elle s'écarte de la vitesse de consigne
L'objet Train contient tous les paramètres du train nécessaires à ses commandes et ses fonctions

**************************************************************/

//////// ETATS DE CIRCULATION ////////////////
enum {              
  INIT=0,
  EXPLORE=1,
  ARRET=2,
  ACCELERE=3,
  PALIER=4,
  RALENTI=5
};

///////// TIMINGS DANS LES ZONES ///////////

#define  ZONE_SAMPLE_TIME     20        // ms
#define  ARRETENGARE          10        // 10 secondes
#define  DISTANCE_MODULE      56        // cm (ZoneB)
#define  VITESSE_MIN          15        //cran

//////////// ZONES /////////////////////////
enum {
  ZGARE1=0,
  ZONEA=1,
  ZONEB=2,
  ZONEC=3,
  ZGARE2=4,
  INCONNU=5
};

const char * nomZone[INCONNU+1]  = {     // noms zones sur LCD
  "Gar1 ",
  "ZonA ",
  "ZonB ",
  "ZonC ",
  "Gar2 ",
  "     "
};


//////////////////////////// OBJET TRAIN ///////////////////////////////

struct Train{
  int adresse_dcc;            // pour la commande DCC
  bool sens1vers2;            // true si gare 1-> gare 2
  bool sens;                  // pour la commande DCC
  bool lumiere;               // pour la commande DCC
  FunctionsState LocoFunctions;    // Current functions of the loco
  int zone;                   // la zone ou il est
  int etat;                   // explore, arret, accelere, palier, ralenti,
  int compteur;               // compteur-décompteur de secondes
  int cran_vitesse;           // cran DCC en cours
  int cran_consigne;          // cran DCC à la consigne
  int consigne;               // en km/H au potar et LCD
  int vmin;                    // cran vitesse minimale
  bool troplent;              // true si Vmin atteinte pendant RALENTI
  int vitesseVraie;           // en km/h
  int acceleration;           // en crans identique dans les 2 sens
  int deceleration1;          // en crans vers gare 1
  int deceleration2;          // en crans vers gare 2
  Train(int);
  void roule();               // fonction pour avancer/reculer
  void stoppe();              // fonction pour s'arreter
  void feux();                // commande des lumieres
  void eteint();              // eteint feux
  void allume();              // allume feux
};

Train::Train(int a) {
  this->adresse_dcc = a;
  this->sens = true;
  this->cran_vitesse = 0;
  this->lumiere = false;
  this->consigne = 40;
  this->cran_consigne = 40;
  this->vmin = 15;
  this->acceleration = 5;
  this->deceleration1 = 5;
  this->deceleration2 = 5;
}

/*********** Fonctions Train ***************/

void Train::roule() { // t1 adresse_dcc vitesse
  DCCpp::setSpeedMain(1, this->adresse_dcc, 128, this->cran_vitesse, this->sens);
}

void Train::stoppe() {
  this->cran_vitesse = 0;
  roule();    
}

void Train::feux() {
  if (this->lumiere) {
    this->LocoFunctions.activate(0);
  } else {
    this->LocoFunctions.inactivate(0);
  }
  DCCpp::setFunctionsMain(2, adresse_dcc, this->LocoFunctions); // registre 1 pour la loco et registre 2 pour la fonction
}

void Train::eteint() {
  this->lumiere=false;
  feux();
}

void Train::allume() {
  this->lumiere=true;
  feux();
}

////////////// GLOBALE gTrain ///////////////

Train gTrain(3);                  //create train default adresse 3

/////////////////ZONEMONITOR ////////////////////
/*
 * ZONEMONITOR ne fait que détecter les entrées dans les zones
 * avec une périodicité de ZONE_SAMPLE_TIME (100 ms)
 * ZoneMonitor::CheckZone est appellé dans la loop principale 
 * le résultat st placé dans la globale gOccupation si checkZone = true
 * pour déclencher les actions correspondant aux changements de zone
 */

struct ZoneMonitor {
  unsigned long zoneTime, zoneTimeAG1, zoneTimeCG2; //zoneTimeA, zoneTimeC 
  int pin1, pinA, pinB, pinC, pin2;
  byte occupation, preoccupation;                 // chaque occupation = un bit (0,1,2,3,4)
  bool zg1, ozg1, za, oza, zb, ozb, zc, ozc, zg2, ozg2;
  ZoneMonitor(int, int, int, int, int);           // creation
  void Init();
  bool checkZone();
};

ZoneMonitor::ZoneMonitor(int pin1, int pina, int pinb, int pinc,int pin2){
    this->pin1=pin1;
    this->pinA=pina;
    this->pinB=pinb;
    this->pinC=pinc;
    this->pin2=pin2;
} 

void ZoneMonitor::Init() {
  pinMode(this->pin1, INPUT_PULLUP);
  pinMode(this->pinA, INPUT_PULLUP);
  pinMode(this->pinB, INPUT_PULLUP);
  pinMode(this->pinC, INPUT_PULLUP);
  pinMode(this->pin2, INPUT_PULLUP);
  zg1 = digitalRead(this->pin1);
  za = digitalRead(this->pinA);
  zb = digitalRead(this->pinB);
  zc = digitalRead(this->pinC);
  zg2 = digitalRead(this->pin2);
  this->ozg1=zg1; 
  this->oza=za; 
  this->ozb=zb; 
  this->ozc=zc; 
  this->ozg2=zg2;
  this->occupation = 0;
  this->preoccupation = 0;
  zoneTime = millis();  
}

bool ZoneMonitor::checkZone() {               // détection de changement d'état
  if (millis()-zoneTime < ZONE_SAMPLE_TIME)   // test periodique (0,1 seconde)
    return(false);                            // trop tôt, plus tard
  zoneTime = millis();   
  this->occupation = 0;
  zg1 = digitalRead(this->pin1);
  if ( zg1 != ozg1){
    ozg1=zg1;  // lecture capteur normalement occupation = LOW
    if (zg1 == LOW) this->occupation = 1;
  } 
  za = digitalRead(this->pinA);
  if ( za != oza ) {
    oza=za;
    if (za == LOW) this->occupation = 2;
  }
  zb = digitalRead(this->pinB);
  if ( zb != ozb ) {
    ozb=zb;
    if (zb == LOW) this->occupation = 4;
    bitWrite(occupation, ZONEB, ozb);
  }
  zc = digitalRead(this->pinC);
  if ( zc != ozc ) {
    ozc=zc;
    if (zc == LOW) this->occupation = 8;
  }
  zg2 = digitalRead(this->pin2);
  if ( zg2 != ozg2 ) {
    ozg2=zg2;
    if (zg2 == LOW) this->occupation = 16;
  }
  if (this->occupation != this->preoccupation && this->occupation != 0) {
    this->preoccupation = this->occupation;
    gOccupation = this->occupation;
    return(true);
  } else {
  return(false);
  }
}

///////////// GLOBALE gzoneMonitor //////////////////

ZoneMonitor gzoneMonitor(PIN_GARE1, PIN_ZONEA, PIN_ZONEB, PIN_ZONEC, PIN_GARE2);

//////////////////////////// AUTOMATE ///////////////////////////////
/* L'automate gère le temps dans chaque état du train (arret, accelere, ralenti
 * 
 */

struct Automate {
  int auto_etat;                  // INIT, EXPLORE, ARRET, ACCELERE, PALIER, RALENTI
  unsigned long arretengare;      // decompteur temps d'arret en gare en ms
  unsigned long timeout; 
  unsigned long timeseconde;      // timecheck
  unsigned long timeLR;           // tempo clignotement Led Rouge
  bool auto_ERR;                  // erreur automate
  bool LRclignote;                // true = Led Rouge clignotante
  // fonctions
  void init();
  void loop();
};

void Automate::init() {
  auto_etat = INIT;
  timeseconde=millis();
  timeLR=millis();
  arretengare = ARRETENGARE;     // tempo d'arret en gare configurable
  auto_ERR = false;
  LRclignote = false;
}

void Automate::loop() {
  if (millis()-timeLR > 500) {  // clignotement LED ROUGE
    if (LRclignote) {  // clignotement LED ROUGE
      digitalWrite(LEDROUGE, !digitalRead(LEDROUGE));
      timeLR = millis();
    } 
  } 
  if (gAutoMan && (millis()-timeseconde > 1000)) { // base de temps 1 seconde
    timeseconde = millis();
    switch (gTrain.etat) {
      case ARRET:                   // decompte temps arret en gare
      gTrain.compteur--;
      if (gTrain.compteur == 0){    // fin arret en gare, redemmarer train
        gTrain.etat = ACCELERE;
        gTrain.cran_vitesse = 10;   // demarrage du train cran 10
        gTrain.roule();
        gTrain.lumiere = true;      // allume lumiere
        gTrain.feux();              // lumière selon 
        #ifdef DEBUG
        Serial.print("ARRET");Serial.println(gTrain.compteur);
        #endif

      }     
      break;
      
      case ACCELERE:
      gTrain.cran_vitesse+=gTrain.acceleration;             
      if (gTrain.cran_vitesse > gTrain.cran_consigne){ // atteinte consigne de vitesse
        gTrain.cran_vitesse = gTrain.cran_consigne;
        gTrain.etat=PALIER;          // etat : palier à vitesse constante
      }
      gTrain.roule();                // commande de vitesse
      #ifdef DEBUG
      Serial.print("ACCELERE");Serial.println(gTrain.cran_vitesse);
      #endif
      break;
      
      case PALIER:
      // aucun changement de comportement
      // attente detection de zone suivante pour ralentissement
      #ifdef DEBUG
        Serial.println("PALIER");
        #endif
      break;
      
      case RALENTI:
      if (gTrain.sens1vers2) {
        gTrain.cran_vitesse-=gTrain.deceleration2;// Vers gare 2
      } else {
        gTrain.cran_vitesse-=gTrain.deceleration1;// Vers gare 1
      }
      if (gTrain.cran_vitesse < gTrain.vmin) {    // atteinte de la consigne minimale 
        gTrain.cran_vitesse = gTrain.vmin;
        gTrain.troplent = true;
        if (gTrain.sens1vers2) {
          gzoneMonitor.zoneTimeCG2 = millis();    // comptage du temps à vitesse MIN de la zone  à la gare 2
        } else {
          gzoneMonitor.zoneTimeAG1 = millis();    // comptage du temps à vitesse MIN de la zone A à la gare 1
        }
      }
      gTrain.roule();
      #ifdef DEBUG
      Serial.print("RALENTI");Serial.println(gTrain.cran_vitesse);
      #endif
      break;
    }
  }
}

/////////// GLOBALE gAutomate /////////////////

Automate gAutomate;   // creation initialisation automate


///////  TRAITEMENT DES DETECTIONS DANS LES ZONES 

void _doGare1() {
  /* Arret train, initialisation du prochain départ, extinction feux,
   * etat = ARRET pour que l'automate decompte le compteur ARRETENGARE
   * quand ce compteur = 0, train redemarre vitesse 10, etat ACCELERE
   * le train entre très vite en ZONEA
   * Venant de ZONEA en etat RALENTI, recalcule du decrement deceleration1
   */
  
  gTrain.stoppe();                        // arrêt
  gTrain.lumiere = false;                 // extinction lumiere
  gTrain.feux();
  gTrain.sens1vers2 = true;               // vers la gare 2
  gTrain.sens = !gTrain.sens;             // inversion du sens DCC
  gTrain.etat = ARRET;
  gTrain.compteur = ARRETENGARE;
  /* Correction du parametre gTrain.deceleration2
   * si la vitesse est supérieure à Vmin (trop vite), on augmente gTrain.deceleration2
   * sinon, trop lent on diminue gTrain.deceleration2
   */
  if (gTrain.etat == RALENTI) {
    // si la vitesse est supérieure à la Vitesse MIN, on augmente gTrain.deceleration1
    if (gTrain.cran_vitesse > gTrain.vmin) {
      gTrain.deceleration1++;
    }
    // si la vmin est atteinte au moins 3 secondes avant la gare on diminue gTrain.deceleration2
    if (gTrain.troplent && (millis()-gzoneMonitor.zoneTimeAG1 > 3000)) {
      gTrain.deceleration2--;
    }
  }
}  

void _doZoneA() {
  /* Venant de GARE1 le train accelere jusqu'à la consigne cran_consigne 
   * etat ACCELERE gere par automate puis etat PALIER
   * puis reste à vitesse constante jusqu'à l'entrée en ZONEB
   * venant de ZONEB, etat = RALENTI ralentissement géré par automate
   * et calcul de vitesse vraie
   */
  gTrain.zone = ZONEA;
  Serial.println(nomZone[gTrain.zone]);
  // calcul de Vitesse seulement dans le sens 2 vers 1 : 1 cm/s <-> 5,76 km/h 
  if (gMesureVitesse && !gTrain.sens1vers2) {  // vers Gare 1
    gcompteurVitesse=millis()-gcompteurVitesse;
    unsigned long VKM = (gEPConfig.dist*5760)/gcompteurVitesse;
    gTrain.vitesseVraie=(int)(VKM);
    gMesureVitesse=false;
    // ajustement de la vitesse gTrain.cran_consigne
    /*if ((gTrain.vitesseVraie > gTrain.consigne + 2) || (gTrain.vitesseVraie < gTrain.consigne - 2)) {
      gTrain.cran_consigne = gTrain.cran_consigne - ((gTrain.vitesseVraie-gTrain.consigne)/2);
      if (gTrain.cran_consigne < gEPConfig.vmin) gTrain.cran_consigne = gEPConfig.vmin;
    }*/
  }
  // ralentissement si automate = PALIER et seulement dans le sens 2 vers 1         
  if (!gTrain.sens1vers2 && (gTrain.etat == PALIER)) {  
    gTrain.etat=RALENTI;                    //etat = ralentissement
    gTrain.troplent=false;
    gAutomate.timeseconde = millis();       // armement de la tempo "seconde"
  }
  // dans le sens 1 vers 2 l'etat est ACCELERE donc rien à faire
}
        
void _doZoneB() {
  /* Initialisation du compteur de vitesse (mesure du temps jusqu'à zone C ou A selon sens)
   * 
   */
  gTrain.zone = ZONEB;
  gcompteurVitesse=millis();                 // lancement compteur de vitesse
  gMesureVitesse=true;   
}
        
void _doZoneC() {
  /* Venant de GARE2 le train accelere jusqu'à la consigne cran_consigne 
   * etat ACCELERE gere par automate puis etat PALIER
   * puis reste à vitesse constante jusqu'à l'entrée en ZONEB
   * venant de ZONEB, etat = RALENTI ralentissement géré par automate
   * et calcul de vitesse vraie
   */
  gTrain.zone = ZONEC;
  // calcul de Vitesse seulement dans le sens G1 vers G2 : 1 cm/s <-> 5,76 km/h 
  if (gMesureVitesse && gTrain.sens1vers2) {  // vers Gare 2
    gcompteurVitesse=millis()-gcompteurVitesse;
    unsigned long VKM = (gEPConfig.dist*5760)/gcompteurVitesse;
    gTrain.vitesseVraie=(int)(VKM);
    gMesureVitesse=false;
    // ajustement de la vitesse gTrain.cran_consigne
    /*if ((gTrain.vitesseVraie > gTrain.consigne + 2) || (gTrain.vitesseVraie < gTrain.consigne - 2)) {
      gTrain.cran_consigne = gTrain.cran_consigne - ((gTrain.vitesseVraie-gTrain.consigne)/2);
      if (gTrain.cran_consigne < gEPConfig.vmin) gTrain.cran_consigne = gEPConfig.vmin;
    }*/
  }
  // ralentissement si automate = PALIER et seulement dans le sens 1 vers 2         
  if (gTrain.sens1vers2 && (gTrain.etat == PALIER)) {  
    gTrain.etat=RALENTI;                    //etat = ralentissement
    gTrain.troplent=false;
    gAutomate.timeseconde = millis();       // armement de la tempo "seconde"
  }
  // dans le sens 1 vers 2 l'etat est ACCELERE donc rien à faire    
}
        
void _doGare2() {
  /* Arret train, initialisation du prochain départ, extinction feux,
   * etat = ARRET pour que l'automate decompte le compteur ARRETENGARE
   * quand ce compteur = 0, train redemarre vitesse 10, etat ACCELERE
   * et le train entre très vite en ZONEB
   * Venant de ZONEB en etat RALENTI recalcule du decrement deceleration2
   */
  gTrain.zone = ZGARE2;
  gTrain.stoppe();                        // arrêt
  gTrain.lumiere = false;                 // extinction lumiere
  gTrain.feux();
  gTrain.sens1vers2 = false;              // vers la gare 1
  gTrain.sens = !gTrain.sens;             // inversion du sens DCC
  gTrain.etat = ARRET;
  gTrain.compteur = ARRETENGARE;
  /* Correction du parametre gTrain.deceleration2
   * si la vitesse est supérieure à Vmin (trop vite), on augmente gTrain.deceleration2
   * sinon, trop lent on diminue gTrain.deceleration2
   */
  if (gTrain.etat == RALENTI) {
    // si la vitesse est supérieure à la Vitesse MIN, on augmente gTrain.deceleration2
    if (gTrain.cran_vitesse > gTrain.vmin) {
      gTrain.deceleration2++;
    }
    // si la vmin est atteinte au moins 3 secondes avant la gare on diminue gTrain.deceleration2
    if (gTrain.troplent && (millis()-gzoneMonitor.zoneTimeCG2 > 3000)) {
      gTrain.deceleration2--;
    }
  }
}
